# TODO Fuel CG
from PySide6.QtWidgets import QWidget, QFormLayout, QLineEdit, QLabel, QCheckBox, QComboBox
AIRCRAFT_TYPES = {
    'Transport': 'transport',
    'General Aviation': 'general_aviation'
}

class aircraft_info(QWidget):
    """
    General aircraft info toolbox
    Use to set
    """

    variable_names = {'Title': 'title',
                      'Design Range': 'design_range',
                      'Cruise Altitude': 'altitude',
                      'Cruise Mach': 'mach',
                      'Maximum Mach': 'max_mach',
                      'Ultimate Load': 'ultimate_load',
                      'Fuel Density': 'rho_fuel'
                      }

    input_types = {'Title': str,
                   'Design Range': float,
                   'Cruise Altitude': float,
                   'Cruise Mach': float,
                   'Maximum Mach': float,
                   'Ultimate Load': float,
                   'Fuel Density': float
                   }

    tool_tips = {'Title': 'Airplane Title',
                   'Design Range': 'Target design range (nmi)',
                   'Cruise Altitude': 'Altitude at cruise (Ft)',
                   'Cruise Mach': 'Mach number at cruise',
                   'Maximum Mach': 'Maximum mach number (~1.05-1.1 time the cruise Mach if unsure)',
                   'Ultimate Load': 'Ultimate load factor (~4.5 if you are unsure)',
                   'Fuel Density': 'Density of the fuel (lbs/gal)'}

    input_fields = {}

    def __init__(self, parent):
        super().__init__(parent)
        self.aircraft = parent.parent.aircraft
        form = QFormLayout(self)
        self.setLayout(form)

        for item in self.variable_names.keys():
            line_edit = QLineEdit()

            if hasattr(self.aircraft, self.variable_names[item]):
                val = str(getattr(self.aircraft, self.variable_names[item]))
            elif hasattr(self.aircraft.mission, self.variable_names[item]):
                val = str(getattr(self.aircraft.mission, self.variable_names[item]))
            else:
                val = ''

            line_edit.textChanged.connect(lambda text, var=item: self.text_changed(var, text))
            line_edit.setText(val)
            self.input_fields[item] = line_edit
            label = QLabel(item)
            label.setToolTip(self.tool_tips[item])
            form.addRow(label, line_edit)

        # Aircraft type select
        self.aircraft_type_select = QComboBox()
        self.aircraft_type_select.addItems(['Transport', 'General Aviation'])
        if self.aircraft.aircraft_type == 'general_aviation':
            self.aircraft_type_select.setCurrentIndex(1)
        self.aircraft_type_select.currentTextChanged.connect(self.aircraft_type_changed)
        form.addRow(QLabel('Aircraft Type'), self.aircraft_type_select)

        # Add lock component weights Checkbox
        self.checkbox = QCheckBox('Lock Component Weights')
        self.checkbox.setChecked(False)
        self.checkbox.stateChanged.connect(self.checkbox_toggled)
        form.addRow(self.checkbox)
        # self.checkbox = QCheckBox()

    def aircraft_type_changed(self, type):
        self.aircraft.aircraft_type = AIRCRAFT_TYPES[type]

    def checkbox_toggled(self, state):
        self.aircraft.lock_component_weights = state == 2

    def text_changed(self, var, text):
        var_name = self.variable_names[var]

        # Validate Input
        try:
            val = self.input_types[var](text)
            self.input_fields[var].setStyleSheet("")

            # Set Value
            if hasattr(self.aircraft, var_name):
                setattr(self.aircraft, var_name, val)
            elif hasattr(self.aircraft.mission, var_name):
                setattr(self.aircraft.mission, var_name, val)

        except ValueError:
            self.input_fields[var].setStyleSheet("border: 1px solid red;")
        except KeyError:
            pass



import math
import os

from PySide6.QtWidgets import QWidget, QFormLayout, QComboBox, QLabel, QLineEdit, QPushButton, QVBoxLayout, QFileDialog, \
    QDialog, QTableWidget, QTableWidgetItem, QHBoxLayout
from PySide6.QtCore import Signal, Qt


class propulsion_input(QWidget):

    n_engines_changed = Signal(int)

    titles = {"n_engines": "Number of Engines",
              "thrust_sea_level": "Maximum Thrust at \nSea Level (lbs)",
              "sfc_sea_level": "Specific Fuel Consumption \nSea Level (/hr)",
              "thrust_cruise": "Maximum Thrust at \nCruise (lbs)",
              "sfc_cruise": "Specific Fuel Consumption \nCruise (/hr)"}

    def __init__(self, parent):
        self.parent = parent
        self.aircraft = parent.parent.aircraft
        self.prop = self.aircraft.propulsion

        super().__init__(parent)
        form = QFormLayout(self)
        self.form_layout = form
        self.setLayout(form)
        prop_type = self.prop.engine_type

        propulsion_type_select = QComboBox(self)
        propulsion_type_select.addItem('Turbofan')
        propulsion_type_select.addItem('Propeller')
        if prop_type == 'propeller':
            propulsion_type_select.setCurrentIndex(1)

        # TODO add some handling here that adjusts the rest of the inputs

        form.addRow(QLabel('Engine Type'), propulsion_type_select)
        self.populate_form_layout(prop_type)

    def populate_form_layout(self, engine_type):

        form = self.form_layout
        self.input_fields = {}
        if engine_type.lower() == 'turbofan':
            # Add file save
            self.file_input = QLineEdit(self.prop.engine_data_file)
            self.browse_button = QPushButton("Browse")
            self.browse_button.clicked.connect(self.open_engine_data_file)
            # self.browse_button.setFixedWidth(80)
            layout = QVBoxLayout()
            layout.addWidget(self.file_input)
            layout.addWidget(self.browse_button)
            layout.setContentsMargins(0, 0, 0, 0)
            form.addRow("Input Data File:", layout)

            for var in self.titles.keys():
                self.input_fields[var] = QLineEdit()
                form.addRow(QLabel(self.titles[var]), self.input_fields[var])
                if getattr(self.prop, var):
                    self.input_fields[var].setText(str(getattr(self.prop, var)))

            # Handle changing the number of engines
            self.input_fields['n_engines'].editingFinished.connect(self.handle_n_engines_changed)

            # Add accept button
            generate_table = QPushButton('Generate engine performance tables')
            generate_table.clicked.connect(self.generate_engine_performance)
            form.addRow(generate_table)

            # Add reset button
            reset_engine = QPushButton('Reset engine parameters')
            reset_engine.clicked.connect(self.reset_engine)
            form.addRow(reset_engine)

    def open_engine_data_file(self):
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Select Data File", "", "Excel Files (*.xlsx);;All Files (*)"
        )
        if file_path:
            self.file_input.setText(file_path)
            self.handle_data_file_changed()

    def handle_data_file_changed(self):
        data_file = self.file_input.text()
        if os.path.exists(data_file):
            self.file_input.setStyleSheet("")
            self.prop.load_data_file(data_file)
        else:
            self.file_input.setStyleSheet("border: 1px solid red;")

    def handle_n_engines_changed(self):

        ac = self.parent.parent.aircraft
        n_engines_old = ac.n_engines
        # Validate Input
        self.input_fields['n_engines'].setStyleSheet("")
        try:
            n_engines = int(self.input_fields['n_engines'].text())
            ac.n_engines = n_engines
        except ValueError:
            self.input_fields['n_engines'].setStyleSheet("border: 1px solid red;")
            return

        n_stations_old = math.ceil(n_engines_old/2)
        n_stations_new = math.ceil(n_engines/2)
        if n_stations_old != n_stations_new:
            self.n_engines_changed.emit(n_engines)

    def reset_engine(self):
        for item in self.input_fields.values():
            item.setText('')

    # TODO fix bug that doesn't let you reset engine data
    def generate_engine_performance(self):
        # Validates input data, generates engine performance table, displays table

        # Validate input
        if not self.validate_input():
            return

        # Generates engine data
        variables = {"n_engines": 2,
                     "thrust_sea_level": None,
                     "sfc_sea_level": None,
                     "thrust_cruise": None,
                     "sfc_cruise": None}

        # Check if inputs are blank
        for item in variables.keys():
            if not self.input_fields[item].text().strip() == '':
                variables[item] = float(self.input_fields[item].text())

        # Generate engine
        if self.prop.engine_type == 'turbofan':
            kwargs = {
                'thrust_sea_level': variables['thrust_sea_level'],
                'thrust_cruise': variables['thrust_cruise'],
                'sfc_sea_level': variables['sfc_sea_level'],
                'sfc_cruise': variables['sfc_cruise'],
                'engine_type': 'turbofan'
            }

        engine = self.aircraft.generate_propulsion(n_engines=variables['n_engines'], **kwargs)

        # Create Dialogue with tables for thrust and sfc
        engine_display = QDialog()
        engine_display.setWindowTitle("Engine Performance Data")

        thrust_table = QTableWidget()
        thrust_table.setRowCount(len(engine.altitude_ref))
        thrust_table.setColumnCount(len(engine.mach_ref))

        thrust_table.setHorizontalHeaderLabels([str(x) for x in engine.mach_ref])
        thrust_table.setVerticalHeaderLabels([str(x) for x in engine.altitude_ref])
        for row in range(len(engine.altitude_ref)):
            for col in range(len(engine.mach_ref)):
                thrust_table.setItem(row, col, QTableWidgetItem(str(engine.thrust_input[row][col])))

        sfc_table = QTableWidget()
        sfc_table.setRowCount(len(engine.altitude_ref))
        sfc_table.setColumnCount(len(engine.mach_ref))

        sfc_table.setHorizontalHeaderLabels([str(x) for x in engine.mach_ref])
        sfc_table.setVerticalHeaderLabels([str(x) for x in engine.altitude_ref])
        for row in range(len(engine.altitude_ref)):
            for col in range(len(engine.mach_ref)):
                sfc_table.setItem(row, col, QTableWidgetItem(str(engine.sfc_input[row][col])))

        main_layout = QVBoxLayout()

        thrust_label = QLabel("Maximum Thrust (lbs)")
        thrust_label.setStyleSheet("font-size: 16pt; font-weight: bold;")
        thrust_label.setAlignment(Qt.AlignCenter)
        # main_layout.addWidget(thrust_label)

        h_layout = QHBoxLayout()
        v_layout = QVBoxLayout()
        v_layout.addStretch()
        v_layout.addWidget(thrust_label)
        h_layout.addWidget(QLabel("Altitude (ft)"))
        mach_label = QLabel('Mach Number')
        mach_label.setAlignment(Qt.AlignCenter)
        v_layout.addWidget(mach_label)
        v_layout.addWidget(thrust_table)
        # v_layout.addStretch()
        h_layout.addLayout(v_layout)
        # h_layout.addStretch()
        main_layout.addLayout(h_layout)


        sfc_label = QLabel("Specific Fuel Consumption (hr^-1)")
        sfc_label.setStyleSheet("font-size: 16pt; font-weight: bold;")
        sfc_label.setAlignment(Qt.AlignCenter)

        h_layout_sfc = QHBoxLayout()
        v_layout_sfc = QVBoxLayout()
        v_layout_sfc.addWidget(sfc_label)
        h_layout_sfc.addWidget(QLabel("Altitude (ft)"))
        mach_label_sfc = QLabel('Mach Number')
        mach_label_sfc.setAlignment(Qt.AlignCenter)
        v_layout_sfc.addWidget(mach_label_sfc)
        v_layout_sfc.addWidget(sfc_table)
        h_layout_sfc.addLayout(v_layout_sfc)
        main_layout.addLayout(h_layout_sfc)

        # Add buttons on bottom
        # Create the three buttons
        save_button = QPushButton("Save Engine Data")

        # Connect actions
        save_button.clicked.connect(self.save_engine_file)
        # exit_button.clicked.connect(self.reject)

        # Create horizontal layout
        button_layout = QHBoxLayout()
        button_layout.addStretch()  # pushes buttons to the right
        button_layout.addWidget(save_button)
        main_layout.addLayout(button_layout)

        main_layout.addStretch()

        engine_display.setLayout(main_layout)
        engine_display.resize(1200, 700)
        engine_display.exec()

    def save_engine_file(self):
        file_path, _ = QFileDialog.getSaveFileName(
            self,
            "Save Engine Data File",
            "data.xlsx",  # default file name
            "Excel Files (*.xlsx)"
        )
        if not file_path:
            return  # user cancelled

            # Ensure file ends with .xlsx
        if not file_path.endswith(".xlsx"):
            file_path += ".xlsx"
        self.prop.write_data_file(file_path)


    def validate_input(self):
        validated = True

        for item in self.input_fields.values():
            try:
                float(item.text())
                item.setStyleSheet("")
            except ValueError:
                if not item.text().strip() == '':
                    item.setStyleSheet("border: 1px solid red;")
                    validated = False
                else:
                    item = None
        return validated





from .wing import Wing
import numpy as np
import logging

logger = logging.getLogger(__name__)

class Wing_advanced(Wing):


    def __init__(self, params):
        # Default Values
        self.title = ''
        self.sweep = 0  # c/4 sweep (rad)
        self.sweep_location = .25  # location of the sweep definition (as a percentage of the chord)
        self.sweep_deg = 0  # c/4 sweep (Deg)
        self.sweep_le = 0  # Leading edge sweep (rad)
        self.sweep_mid = 0  # c/2 sweep (rad)
        self.sweep_te = 0  # Trailing edge sweep (rad)
        self.sweep_quarter_chord = None
        self.area = 0  # Planform surface area (ft^2)
        self.aero_body = True
        self.span = 0  # Wingspan (ft)
        self.cref = 0  # Mean Aerodynamic Chord (ft)
        self.airfoil = ''
        self.aspect_ratio = 0  # Aspect Ratio
        self.taper = 0  # Taper Ratio
        self.cr = 0  # Root Chord
        self.ct = 0  # Tip Chord
        self.Q = 1  # interference factor
        self.laminar_percent = .1  # Percentage of laminar flow
        self.dihedral = 0  # Dihedral angle (rad)
        self.dihedral_deg = 0  # Dihedral angle (deg)
        self.tc = .12  # Airfoil Thickness (At Root)
        self.hc = .008  # Camber
        self.weight_averages = [.8, .2, 0]  # [Raymer, Torenbeek, NASA]
        # self.weight_averages = [.15, .85, 0]
        self.control_surface_ratio = .1
        self.cd0 = 0
        self.laminar_percent = .05
        self.winglet = {}
        self.xc = .35  # x value at mean camber line
        self.avl_sections = []
        self.xle = 0
        self.yle = 0
        self.zle = 0
        super()._load_variables(params)

        self.params = {}
        self.definition_type = "Sections"
        self.avl_sections = []
        self.twist = []
        self.airfoil = []
        self.component_type = 'Wing_Advanced'
        self.sections = []
        self.params = params
        self.component_type = self.__class__.__name__
        for variable_name, variable_value in params.items():
            if hasattr(self, variable_name.lower()):
                setattr(self, variable_name.lower(), variable_value)
        self.create_sections()

    def create_sections(self):
        # Create Sections
        self.sections = []
        first_section = True

        i = 1
        for section_params in self.params['sections']:
            if first_section:
                section_params["xle"] = self.xle
                section_params["yle"] = self.yle
                section_params["zle"] = self.zle
                first_section = False
            else:
                section_params["xle"] = self.sections[-1].xle_tip
                section_params["yle"] = self.sections[-1].yle_tip
                section_params["zle"] = self.sections[-1].zle_tip
                section_params["cr"] = self.sections[-1].ct

            section = wing_section(section_params)
            section.title = f'section__{i}'
            i += 1
            self.sections.append(section)

        # set variables
        self.span = 0
        self.area = 0
        self.cref = 0
        self.s_wet = 0
        for sec in self.sections:
            self.span += sec.span
            self.area += sec.area
            self.cref += sec.cref * sec.area
            self.s_wet += sec.s_wet
        self.cref *= 1 / self.area
        self.taper = self.sections[-1].ct / self.sections[0].cr
        self.aspect_ratio = self.span**2 / self.area
        self.cr = self.sections[0].cr
        self.ct = self.sections[-1].ct
        self.xle = self.sections[0].xle
        self.yle = self.sections[0].yle
        self.zle = self.sections[0].zle
        self.dihedral = np.arctan((self.sections[-1].zle_tip - self.zle) / (self.span*.5))
        self.sweep_location = 0

        self.sweep = np.arctan((self.sections[-1].xle_tip - self.sections[0].xle)/ (self.span/2))
        self.sweep = np.rad2deg(self.sweep)
        self.set_sweep_angle(self.sweep, self.sweep_location)



        if 'airfoils' in self.params:
            self.airfoil = []
            self.twist = []
            for af in self.params['airfoils'].values():
                self.twist.append(af['twist'])
                self.airfoil.append(af['airfoil'])

        # Define avl sections
        if not self.twist:
            self.twist = [0] * (len(self.sections) + 1)
        if not self.airfoil:
            self.airfoil = [None] * (len(self.sections) + 1)

        while len(self.airfoil) < len(self.sections):
            self.airfoil.append(self.airfoil[-1])
        while len(self.twist) < len(self.sections):
            self.twist.append(self.twist[-1])
        self.create_sections()

    def create_avl_sections(self):
        self.avl_sections = [[self.xle, self.yle, self.zle, self.cr, self.twist[0]]]
        i = 1
        for sec in self.sections[1:]:
            self.avl_sections.append([sec.xle, sec.yle, sec.zle, sec.cr, self.twist[i]])
            i += 1

        last_sec = self.sections[-1]
        xle_tip = last_sec.xle + .5 * last_sec.span * np.tan(last_sec.sweep_le)
        yle_tip = last_sec.yle + .5 * last_sec.span
        zle_tip = last_sec.zle + .5 * last_sec.span * np.tan(last_sec.dihedral)

        self.avl_sections.append([xle_tip, yle_tip, zle_tip, last_sec.ct, self.twist[-1]])

    def set_variables(self):
        # Set variables
        for variable_name, variable_value in self.params.items():
            if hasattr(self, variable_name.lower()) and variable_name.lower() != 'sections':
                setattr(self, variable_name.lower(), variable_value)

    def update(self, variable, value, **kwargs):
        section = kwargs.get('section', None)

        if variable in self.params:
            self.params[variable] = value
        else:
            value = float(value)
            if variable == 'span':
                variable = 'length'
                value = value / 2

            if section is not None:
                sec = self.params['sections'][int(section)]
                if variable in sec.keys():
                    sec[variable] = value
            else:
                self.params[variable] = value

        self.__init__(self.params)





class wing_section(Wing):
    input_params = {}

    cr = 0
    ct = 0
    span = 0
    sweep = 0
    sweep_le = 0
    sweep_te = 0
    sweep_location = .25
    dihedral = 0

    xle = 0
    yle = 0
    zle = 0
    xle_tip = 0
    yle_tip = 0
    zle_tip = 0

    area = 0
    taper = 0

    def __init__(self, params):
        self.input_params = params
        for param, value in params.items():
            if hasattr(self, param):
                setattr(self, param.lower(), value)
        self.define_geometry()

    def define_geometry(self):
        params_1 = ['sweep_le', 'sweep_te', 'cr', 'span']   # Defined by leading edge and trailing edge sweep angles
        params_2 = ['cr', 'ct', 'span', 'sweep']            # Defined by root and tip chord

        if all (key in self.input_params for key in params_1):
            # section is defined by leading edge and trailing edge sweep angles
            dy = self.span
            dx_le = self.span * np.sin(np.radians(self.sweep_le))
            dx_te = self.span * np.sin(np.radians(self.sweep_te))
            self.ct = self.cr - dx_le + dx_te
            self.taper = self.ct / self.cr
            self.xle_tip = self.xle + dx_le
            self.area = .5 * self.span * (self.ct + self.cr)
        else:
            if 'length' in self.input_params:
                self.input_params['span'] = self.input_params['length'] * 2
                self.length = self.input_params['length']
            else:
                self.length = self.input_params['span'] / 2
            super().__init__(self.input_params)

        self.xle_tip = self.xle + self.length * np.tan(self.sweep_le)
        self.yle_tip = self.yle + self.length
        self.zle_tip = self.zle + self.length * np.tan(self.dihedral)


from PySide6.QtWidgets import QLineEdit, QLabel, QDialogButtonBox, QWidget, QDialog, QVBoxLayout, QComboBox, \
    QFormLayout, QHBoxLayout, QPushButton
from PySide6.QtCore import Signal

class component_edit(QDialog):
    """ Pop up dialog to edit component parameters"""

    component_info = {"wing": ["Span", "Area", "Taper", "Sweep", "Sweep_location", "Dihedral"],
                      "fuselage": ["Length", "Width", "Height"],
                      "horizontal": ["Span", "Area", "Taper", "Sweep", "Dihedral"],
                      "vertical": ["Span", "Area", "Taper", "Sweep", "Dihedral"],
                      "nacelle": ["Length", "Diameter", "w_engine"]}

    tool_tips = {
        "span": "Tip to tip span (ft)",
        "area": "Total planform area (ft^2)",
        "taper": "Taper ratio, tip chord/root chord",
        "sweep": "Quarter chord sweep angle (deg)",
        "sweep_location": "Location of the sweep angle as a percentage of the chord",
        "dihedral": "Dihedral angle (deg)",
        "length": "Component total length (ft)",
        "width": "Component width in the y direction at thickest point (Ft)",
        "height": "Component height in the z direction at thickest point (Ft)",
        "diameter": "Component diameter at thickest point (Ft)",
        "xle": "X value at component leading edge (ft)",
        "yle": "Y value at component leading edge (ft)",
        "zle": "Z value at component leading edge (ft)",
        "w_engine": "Engine Weight (lbs)"
    }

    variable_labels = {
        "taper": "Taper ratio",
        "ct": "Tip Chord",
        "cr": "Root Chord"
    }

    variable_types = {
        "span": float,
        "area": float,
        "taper": float,
        "sweep": float,
        "sweep_location": float,
        "dihedral": float,
        "length": float,
        "width": float,
        "height": float,
        "diameter": float,
        "xle": float,
        "yle": float,
        "zle": float,
        "w_engine": float,
        "title": str,
        "component_type": str,
        "attachment": str
    }

    common_parameters = ['Xle', 'Yle', 'Zle']
    for comp, param in component_info.items():
        param.extend(common_parameters)
    component_changed = Signal(str)
    title_changed = Signal(str, str)
    custom_component = False

    def __init__(self, component, parent, new_component=False):
        # Set Variables, itialize
        self.new_component = new_component
        # TODO Make "Maintain Aspect Ratio" feature
        self.component = component
        self.title = component
        self.changed_variables = {}
        aircraft = parent.parent.aircraft
        self.aircraft = aircraft
        super().__init__(parent)
        self.init_ui(component)

    def init_ui(self, component):
        self.setWindowTitle("Component Edit")
        aircraft = self.aircraft

        if self.new_component:
            component_type = component.lower()
        else:
            component_type = aircraft.aero_components[component].component_type.lower()
        self.component_type = component_type
        self.layout = QVBoxLayout()

        # Component Type Select
        self.layout.addWidget(QLabel('Component Type'))
        component_type_select = QComboBox()
        component_type_select.setEnabled(False)
        component_type_select.addItems(['Fuselage', 'Wing', 'Horizontal', 'Vertical', 'Nacelle', 'Custom'])
        component_type_select.setCurrentText(component_type.capitalize())
        self.layout.addWidget(component_type_select)

        # Title Edit
        self.layout.addWidget(QLabel('Title'))
        title_edit = QLineEdit()
        if not self.new_component:
            title_edit.setText(str(component.capitalize()))
        title_edit.textChanged.connect(self.handle_title_changed)
        if component.lower() == 'main wing':
            title_edit.setReadOnly(True)
        self.title_edit = title_edit
        self.layout.addWidget(title_edit)

        # Form for all component info
        form = QFormLayout(self)
        form_widget = QWidget()
        form_widget.setLayout(form)

        # Add all component specific information
        self.input_fields = {}
        self.multi_fields = {}
        self.multi_field_values = {}
        self.sub_dicts = {}

        if self.new_component:
            fields = self.component_info[component]
        else:
            fields = aircraft.aero_components[component].params.keys()
            if component not in self.input_fields.keys():
                self.custom_component = True

        for item in fields:
            item = item.lower()
            if item == 'title':
                continue
            # Get Value
            try:
                if self.new_component:
                    val = ''
                else:
                    val = aircraft.aero_components[component].params[item]
                    if item == 'sweep' or item.startswith('dihedral'):
                        val *= 180 / 3.14159
                        val = round(val, 2)
            except KeyError:
                val = ''

            # Set label and tooltips
            if item in self.variable_labels.keys():
                label = QLabel(self.variable_labels[item])
            else:
                label = QLabel(item.title())

            try:
                label.setToolTip(self.tool_tips[item])
            except KeyError:
                pass

            # Add rows
            # Wing sections edit
            if item == "sections" and isinstance(val, list):
                # List of dicts
                container = QWidget()
                layout = QVBoxLayout(container)
                layout.setContentsMargins(0, 0, 0, 0)
                self.sub_dicts[item] = val
                self.section_labels = []

                for idx, section in enumerate(val):
                    row = QWidget()
                    row_layout = QHBoxLayout(row)
                    row_layout.setContentsMargins(0, 0, 0, 0)

                    edit_btn = QPushButton(f"Edit Section {idx + 1}")
                    edit_btn.clicked.connect(lambda _, i=idx: self.edit_section_dialog(i))

                    row_layout.addWidget(edit_btn)
                    layout.addWidget(edit_btn)

                    # self.section_labels.append(summary)
                form.addRow(label, container)

            elif isinstance(val, list):
                container = QWidget()
                layout = QHBoxLayout(container)
                layout.setContentsMargins(0, 0, 0, 0)
                line_edits = []
                for v in val:
                    line_edit = QLineEdit(str(v))
                    line_edit.textChanged.connect(lambda _, var=item: self.multi_field_changed(var))
                    layout.addWidget(line_edit)
                    line_edits.append(line_edit)
                self.multi_fields[item] = line_edits
                self.multi_field_values[item] = val

                form.addRow(label, container)
            else:
                line_edit = QLineEdit()
                line_edit.setText(str(val))
                line_edit.textChanged.connect(lambda text, var=item: self.text_changed(var, text))
                form.addRow(label, line_edit)
                self.input_fields[item] = line_edit

        self.layout.addWidget(form_widget)

        # Ok and Cancel Buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.layout.addWidget(buttons)

        widget = QWidget()
        widget.setLayout(self.layout)

        self.setLayout(self.layout)

    # Stores each changed variable
    def text_changed(self, var, text):
        if hasattr(self.changed_variables, var):
            try:
                self.changed_variables[var] = float(text)
                self.input_fields[var].setStyleSheet("")
            except:
                self.input_fields[var].setStyleSheet("border: 1px solid red;")
        else:
            try:
                self.changed_variables[var] = float(text)
                self.input_fields[var].setStyleSheet("")
            except:
                self.input_fields[var].setStyleSheet("border: 1px solid red;")

    def multi_field_changed(self, item):
        # validate input
        value = []
        for line in self.multi_fields[item]:
            text = line.text().strip()
            try:
                value.append(float(text))
                line.setStyleSheet("")
            except ValueError:
                line.setStyleSheet("border: 1px solid red;")
                validated = False

        # value = [line.text() for line in self.multi_fields[item]]
        self.multi_field_values[item] = value

    def handle_title_changed(self, title):
        self.title = title
        self.changed_variables['title'] = title

    # Closes popup and edits component parameters, signals parent to update graphics
    def accept(self):
        if not self.validate_input():
            return

        for key, value in self.multi_field_values.items():
            self.changed_variables[key] = value
        arg = []
        for key in self.changed_variables.keys():
            if not key == 'title':
                arg.append((self.component, key.lower(), self.changed_variables[key]))

        if arg:
            if not self.new_component:
                self.aircraft.update_component(arg)
                self.component_changed.emit(self.component)
            else:
                params = {}
                for line in arg:
                    params[line[1]] = line[2]
                params['title'] = self.title
                self.aircraft.add_component(self.component_type, params)
                self.component_changed.emit(self.title)
                self.title_changed.emit('', self.title)
                super().accept()
                return
        else:
            self.component_changed.emit(self.component)

        if 'title' in self.changed_variables:
            title = self.title
            self.aircraft.aero_components[title] = self.aircraft.aero_components[self.component]
            self.aircraft.aero_components.pop(self.component)
            self.aircraft.update_component([(title, 'title', title)])
            self.title_changed.emit(self.component, self.title)
        super().accept()

    def validate_input(self):
        validated = True
        for label, line_edit in self.input_fields.items():
            text = line_edit.text().strip()
            if not text:
                line_edit.setStyleSheet("border: 1px solid red;")

            if label in self.variable_types.keys():
                try:
                    self.variable_types[label](text)
                    line_edit.setStyleSheet("")
                except ValueError:
                    line_edit.setStyleSheet("border: 1px solid red;")
                    validated = False

        if 'title' in self.changed_variables and self.new_component:

            if self.changed_variables['title'] in self.aircraft.aero_components:
                self.title_edit.setStyleSheet("border: 1px solid red;")
                validated = False

        return validated

    def edit_section_dialog(self, idx):
        wing = self.aircraft.aero_components[self.component]
        section = wing.sections[idx]
        dlg = QDialog(self)
        dlg.setWindowTitle(f"Edit Section {idx+1}")
        layout = QVBoxLayout(dlg)
        form = QFormLayout()
        edits = {}
        params = section.params

        changed_fields = set()

        def section_text_changed(text):
            changed_fields.add(text)

        for key, val in params.items():
            le = QLineEdit(str(val))
            form.addRow(QLabel(key.title()), le)
            le.textChanged.connect(lambda text, k=key: section_text_changed(k))
            edits[key] = le

        layout.addLayout(form)
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        layout.addWidget(buttons)
        buttons.accepted.connect(dlg.accept)
        buttons.rejected.connect(dlg.reject)

        if dlg.exec_():
            args = []
            for k in changed_fields:
                val = edits[k].text()
                try:
                    v = float(val)
                except ValueError:
                    val = le.text()

                args.append((self.component, k, val))

            if args:
                print(idx)
                self.aircraft.update_component(args, section=idx)
                self.component_changed.emit(self.component)



            # Refresh the whole form/dialog so everything is consistent
            self.component_changed.emit(self.component)  # assumes form-building is in a separate method




from .aircraft import Aircraft
import os
from .components.aerobodies.engine import Engine


def weights_report(aircraft: Aircraft, filename=None):
    """
    Generates a detailed weights report for a given aircraft and writes it to a file.

    :param object aircraft: aircraft which the weights report is for.
    :param str filename: Name of the file that contains the weights report.

    """
    directory = aircraft.output_dir

    if not filename:
        filename = f'{aircraft.file_prefix}_weights.dat'

    with open(os.path.join(directory, filename), 'w') as f:

        f.write(f'Weights Report For Aircraft: {aircraft.title}\n')
        f.write('================================================\n')
        f.write(f'\t\t\t\t\t\t\tWeight (lbs)\n')
        f.write(f'\t\t\t\t\t\t-------------------\n')
        f.write(f'Takeoff Gross Weight: \t\t{aircraft.weight_takeoff:.2f}\n')
        f.write(f'Maximum Takeoff Weight: \t\t{aircraft.weight_max:.2f}\n')
        f.write(f'Operating Empty Weight:\t\t{aircraft.weight_empty:.2f}\n')
        f.write(f'Fuel Weight:\t\t\t\t{aircraft.w_fuel:.2f}\n\n')
        f.write(f'Aircraft Center of Gravity - Empty (ft):\tX:{aircraft.cg_empty[0]:.2f}, Y: {aircraft.cg_empty[1]:.2f}, Z: {aircraft.cg_empty[2]:.2f}\n')
        f.write(
            f'Aircraft Center of Gravity - Full (ft):\t\tX:{aircraft.cg[0]:.2f}, Y: {aircraft.cg[1]:.2f}, Z: {aircraft.cg[2]:.2f}\n\n')
        f.write('================================================\n')
        f.write('          Structural Components\n\n')
        f.write(f'\t\t\t\t\t\tWeight (lbs)\tX cg (ft)\n')
        f.write(f'\t\t\t\t\t\t--------------------------------\n')
        max_variable_name_length = max(len(name) for name in aircraft.aero_components.keys()) + 3
        for comp in aircraft.aero_components.values():
            if isinstance(comp, Engine):
                total_weight = comp.weight  # or comp.weight_total if you have that

            else:
                total_weight = comp.weight
            f.write("{:<{name_width}} {:>{weight_width}.2f} {:>{cg_width}.2f}\n".format(comp.title,
                                                                                  comp.weight,
                                                                                  comp.cg[0],
                                                                   name_width=max_variable_name_length,
                                                                   weight_width=10,
                                                                                  cg_width=12))
            # f.write(f'{comp.title}:\t\t\t\t\t\t{comp.weight}\n')

        f.write('\n================================================\n')
        f.write('             Useful Load \n\n')
        f.write(f'\t\t\t          Weight (lbs)\t\t\tNumber\t\tx cg (ft)\n')
        f.write(f'\t\t            ---------------\t\t-------------   -------\n')
        max_name_width = 15
        f.write("{:<17} {:>11}{:>22}{:>12.2f}\n".format('Passengers',
                                                     aircraft.useful_load.w_passengers,
                                               aircraft.useful_load.n_passengers,
                                                     aircraft.useful_load.cg_passengers[0]))
        f.write("{:<17} {:>11}{:>22}{:>12.2f}\n".format('Pilots', aircraft.useful_load.w_pilots,
                                               aircraft.useful_load.n_pilots,
                                                        aircraft.useful_load.cg_pilots[0]))
        f.write("{:<17} {:>11}{:>22}{:>12.2f}\n".format('Flight Attendants', aircraft.useful_load.w_flight_attendants,
                                               aircraft.useful_load.n_flight_attendants,
                                                        aircraft.useful_load.cg_pilots[0]))
        f.write("{:<17} {:>11}{:>22}{:>12.2f}\n\n".format('Cargo', aircraft.useful_load.w_cargo,
                                                    '', aircraft.useful_load.cg_cargo[0]))

        max_variable_name_length = max(len(name) for name in aircraft.subsystems.components.keys()) + 3

        f.write('================================================\n')
        f.write('          Subsystems\n\n')
        f.write("{:>35}{:>12}\n".format('Weight (lbs)', 'x cg (ft)'))
        f.write("{:>39}\n".format('------------------------------------------'))
        for comp in aircraft.subsystems.components.values():
            f.write("{:<{name_width}} {:>{weight_width}.2f}{:>14.2f}\n".format(comp.title,
                                                                               comp.weight,
                                                                               comp.cg[0],
                                                                   name_width=max_variable_name_length,
                                                                   weight_width=12))

        if aircraft.misc_components:
            f.write('================================================\n')
            f.write('          Misc Weights\n\n')
            f.write("{:>35}{:>12}\n".format('Weight (lbs)', 'x cg (ft)'))
            f.write("{:>39}\n".format('------------------------------------------'))
            for comp in aircraft.misc_components.values():
                f.write("{:<{name_width}} {:>{weight_width}.2f}{:>14.2f}\n".format(comp.title,
                                                                                   comp.weight,
                                                                                   comp.cg[0],
                                                                                   name_width=max_variable_name_length,
                                                                                   weight_width=12))

def mission_profile_report(aircraft, filename=None):

    directory = aircraft.output_dir

    if not filename:
        filename = f'{aircraft.file_prefix}_mission_report.dat'

    with open(os.path.join(directory, filename), 'w') as f:
        f.write(f'Mission profile analysis for Aircraft: {aircraft.title}\n')
        f.write('================================================\n')
        # f.write(f'Maximum Range: {aircraft.}')
        f.write("{:>32} {:>19}\n".format('Range (nmi)', 'Fuel Burnt (lbs)'))
        w_fuel = 0
        for seg in aircraft.mission.mission_profile:
            w_fuel += seg.fuel_burnt
        w_fuel += aircraft.mission.mission_profile[-1].reserve_fuel

        f.write("{:<15} {:>15.2f} {:>17}\n\n".format('Total', aircraft.mission.range, w_fuel))

        for seg in aircraft.mission.mission_profile:
            f.write("{:<15} {:>15.2f} {:>17.2f}\n".format(seg.segment_type, float(seg.range), float(seg.fuel_burnt)))

        try:
            f.write(f"\nReserve and Trap Fuel ({aircraft.mission.mission_profile[-1].wf_reserve*100}%): {aircraft.mission.mission_profile[-1].reserve_fuel:.2f} lbs\n")
        except:
            pass

        i=0
        c1 = 18
        c2 = 15
        for seg in aircraft.mission.mission_profile:
            f.write('\n=========================================================================================\n')
            f.write(f'Segment {i}: {seg.segment_type}\n\n')
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Range (nmi)', seg.range, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Wi (lbs)', seg.wi, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Wn (lbs)', seg.wn, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.6f}\n".format('Weight Fraction', seg.weight_fraction, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n\n".format('Fuel Burnt (lbs)', seg.fuel_burnt, c1=c1, c2=c2))

            if hasattr(seg, 'aoa'):
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Angle of Attack (deg)', seg.aoa, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Lift/Drag', seg.lift_to_drag, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Cl', seg.cl, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Cd', seg.cd, c1=c1, c2=c2))
            if seg.cdw > 0:
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Cdw', seg.cdw, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n\n".format('Cd0', seg.cd0, c1=c1, c2=c2))

            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('altitude', seg.altitude, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Mach', seg.mach, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Velocity (ft/s)', seg.velocity, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n\n".format('Time (min)', seg.time/60, c1=c1, c2=c2))

            if seg.thrust > 0:
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Thrust Required (lbf)', seg.thrust, c1=c1, c2=c2))
            if hasattr(seg, 'max_thrust'):
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Thrust Available (lbf)', seg.max_thrust, c1=c1, c2=c2))
            if seg.sfc > 0:
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('SFC (lb/lbf/hr)', seg.sfc, c1=c1, c2=c2))

            i += 1



from PySide6.QtWidgets import QWidget, QMainWindow, QHBoxLayout, QFileDialog
from .toolbox.toolbox_main import ToolBox
from ..aircraft import Aircraft
from .graphics import graphics


class MainWindow(QMainWindow):
    """ Main Window for WUADS """

    def __init__(self, aircraft):
        super().__init__()
        self.setWindowTitle("WUADS")
        self.resize(1000, 800)

        self.aircraft = aircraft
        self.toolbox = None
        self.graphics = None

        # build menus once
        self._init_menus()

        # build first layout
        self.initiate_window()

    def _init_menus(self):
        """Menus are only created once"""
        menu = self.menuBar()
        file_menu = menu.addMenu('&File')

        load_aircraft = file_menu.addAction('Load')
        load_aircraft.triggered.connect(self.load_config)

        save_aircraft = file_menu.addAction('Save')
        save_aircraft.triggered.connect(self.save_config)

        file_menu.addAction('Close')

    def initiate_window(self):
        """(Re)builds the central widget with toolbox + graphics"""

        # clear the old central widget if it exists
        old_widget = self.centralWidget()
        if old_widget is not None:
            self.setCentralWidget(None)
            old_widget.deleteLater()
            self.toolbox = None
            self.graphics = None

        # rebuild
        layout = QHBoxLayout()


        self.toolbox = ToolBox(self)
        layout.addWidget(self.toolbox, 1)

        self.graphics = graphics(self)
        layout.addWidget(self.graphics, 3)

        # rewire signals
        self.toolbox.component_changed.connect(self.graphics.update_component)
        self.toolbox.component_selected.connect(self.graphics.handleComponentSelected)
        self.toolbox.component_renamed.connect(self.graphics.handleComponentRenamed)

        # draw new aircraft
        self.graphics.plot_aircraft(self.aircraft)

        container = QWidget()
        container.setLayout(layout)
        self.setCentralWidget(container)

    def load_config(self, *args):
        """Reload configuration + rebuild UI"""
        file_dialog = QFileDialog(self)
        file_dialog.setNameFilter("*.yml *.yaml")
        if file_dialog.exec():
            config_file = file_dialog.selectedFiles()
            if config_file:
                self.aircraft = Aircraft(str(config_file[0]))
                self.initiate_window()


    def save_config(self, *args):
        file_dialog = QFileDialog.getSaveFileName(
            parent=self,
            caption='Save Aircraft Configuration File',
            filter='*.yml *.yaml'
        )
        file_name = file_dialog[0]
        self.aircraft.write_config_file(file_name)



import numpy as np


# Calculates atmospheric conditions and flight conditions with a given height and mach number
class FlightConditions:
    """
    Calculates atmospheric and flight conditions for a given altitude and Mach number.

    """

    def __init__(self, altitude, mach):
        """
        Initializes the FlightConditions object with the specified altitude and Mach number.

        :param float altitude: altitude (ft)
        :param float mach: mach number
        """

        self.pressure = 0
        self.temperature = 0
        self.rho0 = 0
        self.mu = 0
        self.a = 0
        self.velocity = 0
        self.mach = 0
        self.q = 0
        self.re = 0
        self.altitude = 0
        self.gamma = 1.4

        # Constants
        g = 32.2  # gravity
        t0 = 519  # Rankine
        p0 = 2116  # lb/ft^2
        rho0 = .0023769
        a0 = 1116  # ft/s
        L = 0.003575  # temp lapse rate (R/ft)
        R = 1716.5  # Gas Constant
        gamma = 1.4

        t = t0 - L * altitude
        theta = t / t0
        delta = theta ** (g / (L * R))
        p = delta * p0
        sigma = delta / theta

        # Stratosphere Correction
        if delta < 0.224:
            h0 = 36000  # Troposphere Ceiling
            beta = 20790
            dh = altitude - 36000
            theta = .752
            delta = theta ** (g / (L * R)) * np.exp(-dh / beta)
            sigma = delta / theta
            t = t0 * theta

        # Sutherland's law (calculates in metric, converts back to imperial
        c1 = 1.458e-6
        S = 110.4
        tref = t * 5 / 9  # Rankine to Kelvin
        self.mu = c1 * (tref ** 1.5) / (tref + S)
        self.mu *= .0208854  # metric to imperial

        self.temperature = t
        self.pressure = delta * p0
        self.rho = sigma * rho0
        self.a = np.sqrt(gamma * R * t)

        self.velocity = mach * self.a
        self.mach = mach
        self.q = .5 * self.rho * self.velocity ** 2
        self.altitude = altitude



from PySide6.QtWidgets import QDialog, QVBoxLayout, QLabel, QLineEdit, QFormLayout, QWidget, QDialogButtonBox
from ...components.component import Component

subcomponent_types = {'Nose Landing Gear': 'nose_landing_gear',
                      'Main Landing Gear': 'main_landing_gear',
                      'Air Conditioning': 'air_conditioning' ,
                      'Anti Icing': 'anti_ice',
                      "Auxiliary Power Unit (APU)": 'apu',
                      'Avionics': 'avionics',
                      'Electronics': 'electronics',
                      'Flight Controls': 'flight_controls',
                      'Furnishings': 'furnishings',
                      'Hydraulics': 'hydraulics',
                      'Instruments': 'instruments'}

class subcomponent_edit(QDialog):

    component_info = {'Nose Landing Gear': {'Number of Wheels': 'n_nose_wheels'},
                      'Main Landing Gear': {'Number of Wheels': 'n_main_wheels'},
                      'Air Conditioning': {},
                      'Anti Icing': {},
                      "Auxiliary Power Unit (APU)": {},
                      'Avionics': {'Uninstalled Avionics Weight': 'w_avionics'},
                      'Electronics': {},
                      'Flight Controls': {},
                      'Furnishings': {},
                      'Hydraulics': {},
                      'Instruments': {}}

    cg = [0.0, 0.0, 0.0]

    def __init__(self, component, parent):
        # Initiate Window
        super().__init__(parent)
        self.setWindowTitle('Subsystem Edit')
        self.aircraft = parent.parent.aircraft
        self.title = subcomponent_types[component]
        if self.title in self.aircraft.subsystems.components:
            self.component = self.aircraft.subsystems.components[self.title]
            self.cg = self.component.cg
        else:
            fuse = self.aircraft.aero_components['Fuselage']
            self.cg = cg_default(self.title, fuse)

        # Create Layout
        self.layout = QVBoxLayout()

        # Title Display
        self.layout.addWidget(QLabel('Subsystem Component'))
        self.title_display = QLineEdit()
        self.title_display.setText(component)
        self.title_display.setReadOnly(True)
        self.layout.addWidget(self.title_display)

        # form layout
        form = QFormLayout(self)
        form_widget = QWidget()
        form_widget.setLayout(form)
        self.layout.addWidget(form_widget)

        self.x_edit = QLineEdit()
        self.x_edit.setText(str(self.cg[0]))
        form.addRow(QLabel('X'), self.x_edit)
        self.y_edit = QLineEdit()
        self.y_edit.setText(str(self.cg[1]))
        form.addRow(QLabel('Y'), self.y_edit)
        self.z_edit = QLineEdit()
        self.z_edit.setText(str(self.cg[2]))
        form.addRow(QLabel('Z'), self.z_edit)

        # Buttons
        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self.layout.addWidget(buttons)

        # Set Layout
        widget = QWidget()
        widget.setLayout(self.layout)
        self.setLayout(self.layout)

    def accept(self):

        if not self.validate_input():
            return
        self.cg = [float(self.x_edit.text()), float(self.y_edit.text()), float(self.z_edit.text())]
        self.aircraft.subsystems.components[self.title] = Component({'title': self.title, 'cg': self.cg})
        self.aircraft.set_weight()
        super().accept()


    def validate_input(self):
        # Validate
        valid = True
        try:
            x = float(self.x_edit.text())
        except ValueError:
            self.x_edit.setStyleSheet("border: 1px solid red;")
            valid = False
        try:
            y = float(self.y_edit.text())
        except ValueError:
            self.y_edit.setStyleSheet("border: 1px solid red;")
            valid = False
        try:
            z = float(self.z_edit.text())
        except ValueError:
            self.z_edit.setStyleSheet("border: 1px solid red;")
            valid = False
        return valid

def cg_default(component, fuse):
    r = fuse.height / 2
    l = fuse.length
    component_cg = {'nose_landing_gear': [15, 0, -6],
                        'main_landing_gear': [69.3, 0, -6],
                        'air_conditioning': [62, 0, 0],
                        'anti_ice': [67, 0, 0],
                        'apu': [121, 0, 0],
                        'avionics': [9, 0, 0],
                        'electronics': [12, 0, 0],
                        'flight_controls': [0, 0, 0],
                        'furnishings': [56, 0, 0],
                        'hydraulics': [62, 0, 0],
                        'instruments': [9, 0, 0],
                      }
    cg = component_cg[component]
    cg[0] *= l / 123.31
    cg[1] *= r / 13.17
    cg[2] *= r / 13.17

    return cg



from ..component import PhysicalComponent
import numpy as np


class Fuselage(PhysicalComponent):
    """
    Class containing all methods and variables for fuselage.
    Used for ellipse shaped main fuselages.
    """

    def __init__(self, params):
        """
        Initializes fuselage with given parameters.

        :param <dict> params: list of parameters to edit
        """

        # Default Values
        self.length = 0          # Length Tip to Nose (ft)
        self.width = 0           # Width (ft)
        self.height = 0          # Height (ft)
        self.a_max = 0           # Cross-sectional area at thickest point (ft^2)
        self.diameter = 0        # Average Diameter (Ft)
        self.fineness_ratio = 0  # Form Factor
        self.s_wet = 0


        self.Q = 1.1             # Interference Factor - Landing gear and wing correction
        self.laminar_percent = .05  # Laminar flow percentage
        self.fuse_mounted_lg = True  # Whether the landing gear is mounted

        self.weight_averages = [.5, 0, .5]  # [Raymer, Torenbeek, NASA] - weighted averages used for weight estimation

        super().__init__(params)

        # Calculate class variables
        self.diameter = .5 * (self.width + self.height)
        self.a_max = .25 * np.pi * self.diameter ** 2
        if self.diameter == 0 or self.length == 0:
            return

        self.fineness_ratio = self.length / self.diameter
        f = self.fineness_ratio
        # Torenbeek 1988 wetted area estimation
        self.s_wet = np.pi * self.diameter * self.length * (1 - 2 / f) ** (2 / 3) * (1 + 1 / (f ** 2))

    # def parasite_drag(self, flight_conditions, sref):
    #     """
    #     Calculates fuselage drag coefficient using method from gotten et. al.
    #     https://doi.org/10.2514/1.C036032
    #
    #     :param flight_conditions: flight conditions object at cruise
    #     :param float sref: aircraft reference area
    #     """
    #     if self.diameter == 0 or self.length == 0:
    #         self.cd0 = 0
    #         self.cdw = 0
    #         return 0
    #
    #
    #     l_char = self.length
    #     d = self.diameter
    #     w = self.width
    #     cs1 = -.825885 * (d / w) ** .411795 + 4.0001
    #     cs2 = -.340977 * (d / w) ** 7.54327 - 2.27920
    #     cs3 = -.013846 * (d / w) ** 1.34253 + 1.11029
    #     form_factor = cs1 * (l_char / d) ** cs2 + cs3
    #
    #     re = flight_conditions.rho * flight_conditions.velocity * l_char / flight_conditions.mu
    #     cf = (1 / (3.46 * np.log10(re) - 5.6)) ** 2
    #     if sref == 0:
    #         return
    #     self.cd0 = cf * form_factor * self.Q * self.s_wet / sref
    def parasite_drag(self, flight_conditions, sref, aircraft):
        """
        Calculates fuselage drag coefficient using method from gotten et. al.
        https://doi.org/10.2514/1.C036032

        :param flight_conditions: flight conditions object at cruise
        :param float sref: aircraft reference area
        """
        sref = aircraft.sref
        if self.diameter == 0 or self.length == 0:
            self.cd0 = 0
            self.cdw = 0
            return 0

        if aircraft.aircraft_type == 'transport':
            l_char = self.length
            d = self.diameter
            w = self.width
            cs1 = -.825885 * (d / w) ** .411795 + 4.0001
            cs2 = -.340977 * (d / w) ** 7.54327 - 2.27920
            cs3 = -.013846 * (d / w) ** 1.34253 + 1.11029
            form_factor = cs1 * (l_char / d) ** cs2 + cs3
        elif aircraft.aircraft_type == 'general_aviation':
            # Raymer method instead
            l_char = self.length  # Fuselage length
            d = self.diameter  # Fuselage diameter

            # Raymer fuselage form factor (Eq. 12.31)
            FF_raymer = 1 + (60 / (l_char / d) ** 3) + 0.0025 * (l_char / d)

            # Apply GA correction factor for bluffness, roughness, etc.
            correction_factor = 1.4  # 30% increase for light aircraft
            form_factor = FF_raymer * correction_factor

        re = flight_conditions.rho * flight_conditions.velocity * l_char / flight_conditions.mu
        cf = (1 / (3.46 * np.log10(re) - 5.6)) ** 2

        #change to equivalent skin friction
        #cf = .0055
        if sref == 0:
            return
        self.cd0 = cf * form_factor * self.Q * self.s_wet / sref


    def set_weight(self, aircraft, wdg):
        if self.length == 0 or self.diameter == 0:
            self.weight = 0
            return 0
        else:
            # if aircraft.aircraft_type == 'general_aviation':
            #     self.weight_averages = [0, 0, 1]
            # super().set_weight(aircraft, wdg)
            self.weight = self.nasa_weight(aircraft, wdg)
            self.set_cg()
            return self.weight

    def raymer_weight(self, aircraft, wdg):
        """
        Calculates fuselage weight using raymer method

        :param object aircraft: <aircraft object> aircraft which the fuselage belongs to
        :param int wdg: takeoff gross weight (lbs)

        :return: Raymer weight (lbs)
        :rtype: float
        """
        if aircraft.aircraft_type == 'transport':
            bw = aircraft.aero_components['Main Wing'].span
            tw = aircraft.aero_components['Main Wing'].taper
            sweepw = aircraft.aero_components['Main Wing'].sweep

            if 'kdoor' in self.params:
                kdoor = self.params['kdoor']
            else:
                kdoor = 1.06

            if self.fuse_mounted_lg:
                klg = 1.12
            else:
                klg = 1.0

            kws = .75 * ((1 + 2 * tw) / (1 + tw)) * (bw * np.tan(sweepw / self.length))
            self.weight_raymer = 0.328 * kdoor * klg * (
                        wdg * aircraft.ultimate_load) ** .5 * self.length ** .25 * \
                                 self.s_wet ** .302 * (1 + kws) ** .04 * (self.fineness_ratio) ** .1

            return self.weight_raymer

        elif aircraft.aircraft_type == 'general_aviation':

            tail_name = None
            for name, comp in aircraft.aero_components.items():
                if comp.component_type == 'horizontal':
                    tail_name = name

            if not tail_name:
                lt = (self.xle + self.length) - (aircraft.aero_components['Main Wing'].xle + .25 * aircraft.aero_components['Main Wing'].cr)
            else:
                lt = (aircraft.aero_components[tail_name].xle + .25 * aircraft.aero_components[tail_name].cr) - \
                     (aircraft.aero_components['Main Wing'].xle + .25 * aircraft.aero_components['Main Wing'].cr)

            Wpress = 0
            a = .052 * self.s_wet ** 1.086
            b = (aircraft.ultimate_load * wdg) **.414 * lt ** -.051
            c = (self.fineness_ratio) ** -.072
            d = aircraft.cruise_conditions.q ** .241
            self.weight_raymer = a * b * c * d + Wpress

            return self.weight_raymer

    def torenbeek_weight(self, aircraft, wdg):
        """
        Calculates fuselage weight using Torenbeek method.

        :param: object aircraft: <aircraft object> aircraft which the fuselage belongs to
        :param: int wdg: takeoff gross weight (lbs)

        :return: Torenbeek weight (lbs)
        :rtype: float
        """

        # Adjust constants from inputs
        kf = 1
        if self.fuse_mounted_lg:
            kf *= 1.07

        if 'pressurized_fuse' in self.params:
            if self.params['pressurized_fuse']:
                kf *= 1.08
        else:
            kf *= 1.08

        if 'cargo_floor' in self.params:
            if self.params['cargo_floor']:
                kf *= 1.1

        # estimates tail ac as the end of the fuselage, should be fairly accurate except for unconventional
        # configurations
        acw = aircraft.aero_components['Main Wing'].xle + .25 * aircraft.aero_components['Main Wing'].cr
        lh = self.length - acw

        # TODO Figure out how to implement divespeed
        dive_speed = aircraft.cruise_conditions.v * .592484
        self.weight_torenbeek = .021 * kf * ((dive_speed * lh) / (self.width + self.height)) ** .5 * self.s_wet ** 1.2
        return self.weight_torenbeek

    def nasa_weight(self, aircraft, wdg):
        """
        Calculates fuselage weight using NASA FLOPS method.

        :param: object aircraft: <aircraft object> aircraft which the fuselage belongs to.
        :param: int wdg: takeoff gross weight (lbs)

        :return: NASA FLOPS weight (lbs)
        :rtype: float
        """
        if aircraft.aircraft_type == 'transport':
            # number of fuselage mounted engines
            n_fuse_en = 0
            for comp in aircraft.aero_components.values():
                if comp.component_type == 'Nacelle' and comp.attachment == self.title:
                    if comp.yle == 0:
                        n_fuse_en += 1
                    else:
                        n_fuse_en += 2

            if 'cargo_floor' in self.params and self.params['cargo_floor']:
                cargf = 1.0
            else:
                cargf = 0

            self.weight_nasa = 1.35 * (self.length * self.height) ** 1.28 * (1 + .05 * n_fuse_en) * \
                               (1 + .38 * cargf) * 1

        elif aircraft.aircraft_type == 'general_aviation':
            a = .052 * self.s_wet ** 1.086
            b =(aircraft.ultimate_load * wdg) ** .177
            c = aircraft.cruise_conditions.q ** .241
            self.weight_nasa =  a * b * c



        return self.weight_nasa

    def set_cg(self):
        """
        Sets center of gravity and moments of inertia for fuselage.
        """
        self.cg = [self.length * .5 + self.xle, 0, 0]
        super().set_cg()

    def update(self, variable, value):
        """
        Allows for modifying a specific fuselage property (such as `width` or `height`)
        and recalculates derived attributes like `diameter` based on the updated value.

        :param str variable: The name of the fuselage property to be updated (e.g., 'width', 'height').
        :param float value: The new value to set for the specified property.
        :param maintain_aspect_ratio: Indicates whether to maintain the aspect ratio between `width` and `height`.
        """
        super().update(variable, value)
        self.diameter = .5 * (self.width + self.height)


# src/wuads/main.py

import sys
import argparse
from pathlib import Path
from importlib import resources
from PySide6.QtWidgets import QApplication
from .main_window import MainWindow
from ..aircraft import Aircraft  # Handles loading config internally

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "config",
        nargs="?",  # Makes it optional
        help="Path to input configuration .yml file"
    )
    args = parser.parse_args()

    if args.config:
        config_path = Path(args.config)
    else:
        try:
            # Works when installed as a package
            default_path = resources.files("WUADS.assets").joinpath("737-800.yml")
        except (ModuleNotFoundError, AttributeError):
            # Fallback for running from cloned repo
            default_path = Path(__file__).parent.parent / "assets" / "737-800.yml"
        config_path = default_path

    aircraft = Aircraft(config_path)  # Pass path directly to Aircraft

    app = QApplication(sys.argv)
    window = MainWindow(aircraft=aircraft)
    window.show()
    sys.exit(app.exec())

if __name__ == "__main__":
    main()


from .mission_segments import takeoff, climb, cruise, descent, landing, loiter, weight_drop
import logging

logger = logging.getLogger(__name__)

MISSION_KEYS = {
    'takeoff': takeoff,
    'climb': climb,
    'cruise': cruise,
    'descent': descent,
    'landing': landing,
    'loiter': loiter,
    'weight_drop': weight_drop
}

class Mission:
    """
    Class which contains information to the mission requirements and profile of an aircraft
    """

    def __init__(self, aircraft):

        # Default values
        self.w_fuel = 0
        self.n_passengers = 0
        self.n_pilots = 0
        self.n_flight_attendants = 0
        self.mach = 0
        self.max_mach = 0
        self.altitude = 0
        self.design_range = 0
        self.range = 0

        self.rho_fuel = 6.8
        self.ultimate_load = 0

        self.mission_profile = []

        self.aircraft = aircraft
        self.max_mach = self.mach * 1.025

        self.takeoff_data = None
        self.climb_data = None
        self.cruise_data = None
        self.descent_data = None
        self.landing_data = None


    def generate_mission_profile(self, params):

        # Note: this should only really run on initialization
        if not params:
            self.mission_profile = [
                takeoff(thrust_setting=75, time=30),
                climb(aircraft=self.aircraft, start_velocity=150, end_velocity=200, start_altitude=0, end_altitude=10000,
                      best_climb=False),
                cruise(aircraft=self.aircraft, mach=0.85, altitude=35000, find_range=True),
                descent(weight_fraction=0.95),
                landing(weight_fraction=0.9, reserve_fuel=0.1)
            ]
        else:
            for name, seg in params.items():
                seg_class = MISSION_KEYS.get(seg['segment_type'])
                self.mission_profile.append(seg_class(aircraft=self.aircraft, title=name, **seg))


    def run_case(self, mute_output=False):
        """
        Runs mission profile calculations using the user-defined mission profile.
        Assumes self.mission_profile is already populated with valid mission segments.
        """
        aircraft = self.aircraft
        mission_profile = self.mission_profile  # use the existing mission profile defined by user
        wi = aircraft.weight_takeoff


        if not mute_output:
            logger.info("Generating mission profile...")

        # Forward loop: compute weight fractions and range up to the find_range segment
        seg_findrange = None
        for i, seg in enumerate(mission_profile):
            if seg.find_range:
                indx_findrange = i
                seg_findrange = seg
                break
            else:
                if not mute_output:
                    logger.info(f"Analyzing conditions for mission segment {seg.title}")
                seg.breguet_range(aircraft, wi=wi)
                wi *= seg.weight_fraction

        if seg_findrange is None:
            pass
        else:
            # Backward loop: compute wi for segments after find_range
            wn = aircraft.weight_takeoff - aircraft.useful_load.w_fuel
            for seg in reversed(mission_profile):
                if seg.find_range:
                    break
                else:
                    if not mute_output:
                        logger.info(f"Analyzing conditions for mission segment {seg.title}")
                    seg.breguet_range(aircraft, wn=wn)
                    wn = seg.wi

            if not mute_output:
                logger.info(f"Finding maximum range")
            seg_findrange.set_range(aircraft, wi=wi, wn=wn)

        # Sum total mission range
        max_range = sum(seg.range for seg in mission_profile)
        if not mute_output:
            logger.info(f"Analysis complete, maximum range is {max_range}")
        aircraft.range = max_range
        self.mission_profile = mission_profile  # store updated mission profile
        self.range = max_range



import os
import subprocess
import sys
import logging

logger = logging.getLogger(__name__)
# Writes .avl file readable by avl
# Writes avl geometry file
def AVL_input(ac, w, mach=None):
    if mach == None:
        mach = ac.cruise_conditions.mach

    output_dir = ac.output_dir
    plane_file = os.path.join(output_dir, f'{ac.file_prefix}_plane.avl')
    mass_file = os.path.join(output_dir, f'{ac.file_prefix}_mass.mass')

    # Writes a .avl file that's readable by the program
    fid = open(plane_file, 'w')

    fid.write('AVL Geometry\n\n')
    fid.write('#Mach\n')
    fid.write(f'{mach}\n\n')
    fid.write('#IYsym   IZsym   Zsym\n')
    fid.write('0       0       0\n')
    fid.write('#Sref    Cref    Bref\n')
    fid.write("{0}   {1}   {2}\n\n".format(ac.sref, ac.aero_components['Main Wing'].cref, ac.aero_components['Main Wing'].span))
    fid.write("#Xref    Yref    Zref\n")
    fid.write('{0}     {1}     {2}\n'.format(ac.cg[0], ac.cg[1], ac.cg[2]))
    fid.write('#--------------------------------------------------\n')

    components = {}
    components['Main Wing'] = ac.aero_components['Main Wing']
    for comp in ac.aero_components.keys():
        if not comp == 'Main Wing':
            components[comp] = ac.aero_components[comp]

    for comp in components.values():
        if comp.aero_body:
            fid.write('SURFACE\n')
            fid.write('{0}\n\n'.format(comp.title))
            fid.write('!Nchordwise  Cspace  Nspanwise  Sspace\n')
            fid.write('12           2.0     26         -1.5\n')
            if 'wing' in comp.component_type.lower():
                fid.write('Component\n')
                fid.write('1\n\n')
                fid.write('Angle\n2 \n\n')

            if 'vertical' not in comp.component_type:
                fid.write('YDUPLICATE\n')
                fid.write('0\n\n')

            fid.write('SCALE\n')
            fid.write('1   1   1\n\n')

            # Sections
            for j in range(len(comp.avl_sections)):
                fid.write('Section\n')
                fid.write('#Xle    Yle    Zle     Chord   Ainc  Nspanwise  Sspace\n')
                fid.write('{0}  {1}  {2}  {3}  {4}\n'.format(comp.avl_sections[j][0], comp.avl_sections[j][1],
                                                             comp.avl_sections[j][2], comp.avl_sections[j][3],
                                                             comp.avl_sections[j][4]))

                if 'horizontal' in comp.component_type.casefold():
                    fid.write('Control\n')
                    fid.write('Elevator 1.0 {0} 0. 1. 0. 1\n'.format(comp.control_surface_ratio))
                if len(comp.airfoil) == 1:
                    fid.write('AFILE\n')
                    fid.write('{0}\n\n'.format(comp.airfoil))
                elif len(comp.airfoil) > 1:
                    fid.write('AFILE\n')
                    fid.write(f'{comp.airfoil[j]}\n\n')
                else:
                    fid.write('\n')

            fid.write('#--------------------------------------------------\n')

    fid.close()

    # Write Mass File
    fid = open(mass_file, 'w')
    fid.write('Lunit = 3.048000e-01 m\n')
    fid.write('Munit = 4.535000e-01 kg\n')
    fid.write('Tunit = 1 s\n\n')
    fid.write('{0}  {1}  {2}  {3}  {4}  {5}  {6}\n'.format(w, ac.cg[0], ac.cg[1], ac.cg[2], ac.inertia[0], ac.inertia[1], ac.inertia[2]))
    fid.close()


# runs specified case and saves results in derivs.st file
def run_AVL(fc, ac, cd0=None, cdw=None):

    if cd0 is None:
        cd0 = ac.cd0
    if cdw is None:
        cdw = ac.cdw

    output_dir = ac.output_dir
    derivs_file = os.path.join(output_dir, 'derivs')

    # Conversion Factors
    slg2kgm = 515.379
    ft2m = 0.3048

    # %% Inputs
    geom_file = os.path.join(output_dir, f'{ac.file_prefix}_plane.avl')
    mass_file = os.path.join(output_dir, f'{ac.file_prefix}_mass.mass')
    M = fc.mach
    Cd0 = cd0 + cdw

    a = fc.a * ft2m
    V = M * a

    if os.path.exists(derivs_file):
        os.remove(derivs_file)

    commands = (f"LOAD {geom_file}\n"
                f"Mass {mass_file}\n"
                f"MSET\n"
                "0\n"
                "Oper\n"
                "C1\n"
                "G 9.81\n"
                f"D {fc.rho * slg2kgm}\n"
                f"V {V}\n\n"
                "D1 PM 0\n"
                "M\n"
                f"MN {M}\n"
                f"CD {Cd0}\n\n"
                "x\n"
                "st\n"
                f"{derivs_file}\n\n"
                "Quit\n\n")
    # Run avl

    try:
        process = subprocess.run(['avl'],
                                 input=commands.encode(),
                                  stdout=subprocess.DEVNULL,
                                  )
    except FileNotFoundError:
        logger.error('AVL.exe file not found, please add to working directory or add avl to environment variables')
        sys.exit(1)
        return

# Imports lift and drag coefficients from output derivs.st file
def import_coefficients(ac, seg):
    derivs_file = os.path.join(ac.output_dir, 'derivs')

    try:
        with open(derivs_file, 'r') as fid:
            derivs = fid.readlines()[23:25]
    except FileNotFoundError:
        logger.error('AVL failed to achieve trim conditions')
        sys.exit(1)

    cl = []
    for t in derivs[0].split():
        try:
            cl.append(float(t))
        except ValueError:
            pass

    cd = []
    for t in derivs[1].split():
        try:
            cd.append(float(t))
        except ValueError:
            pass

    with open(derivs_file, 'r') as fid:
        aoa_dat = fid.readlines()[15]
    seg.cl = cl[0]
    seg.cd = cd[0]
    aoa = aoa_dat.split()
    seg.aoa = float(aoa[2])
    return cl[0], cd[0]

def mission_profile_report(aircraft, filename):
    with open(filename, 'w') as f:
        f.write(f'Mission profile analysis for Aircraft: {aircraft.title}\n')
        f.write('================================================\n')
        # f.write(f'Maximum Range: {aircraft.}')
        f.write("{:>32} {:>19}\n".format('Range (nmi)', 'Fuel Burnt (lbs)'))
        f.write("{:<15} {:>15.2f} {:>17}\n\n".format('Total', aircraft.range, aircraft.UsefulLoad.w_fuel))

        for seg in aircraft.mission_profile:
            f.write("{:<15} {:>15.2f} {:>17.2f}\n".format(seg.segment_type, float(seg.range), float(seg.fuel_burnt)))

        f.write(f"\nReserve and Trap Fuel ({aircraft.mission_profile[-1].wf_reserve*100}%): {aircraft.mission_profile[-1].reserve_fuel:.2f} lbs\n")

        i=0
        c1 = 18
        c2 = 15
        for seg in aircraft.mission_profile:
            f.write('\n=========================================================================================\n')
            f.write(f'Segment {i}: {seg.segment_type}\n\n')
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Range (nmi)', seg.range, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Wi (lbs)', seg.wi, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Wn (lbs)', seg.wn, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.6f}\n".format('Weight Fraction', seg.weight_fraction, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.2f}\n\n".format('Fuel Burnt (lbs)', seg.fuel_burnt, c1=c1, c2=c2))

            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('Lift/Drag', seg.lift_to_drag, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Cl', seg.cl, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n\n".format('Cd', seg.cd, c1=c1, c2=c2))

            f.write("\t{:<{c1}}{:>{c2}.2f}\n".format('altitude', seg.altitude, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Mach', seg.mach, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Velocity (m/s)', seg.velocity, c1=c1, c2=c2))
            f.write("\t{:<{c1}}{:>{c2}.4f}\n\n".format('Time (min)', seg.time/60, c1=c1, c2=c2))

            if seg.thrust > 0:
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Thrust Required (lbf)', seg.thrust, c1=c1, c2=c2))
            if hasattr(seg, 'max_thrust'):
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('Thrust Available (lbf)', seg.max_thrust, c1=c1, c2=c2))
            if seg.sfc > 0:
                f.write("\t{:<{c1}}{:>{c2}.4f}\n".format('SFC (lb/lbf/hr)', seg.sfc, c1=c1, c2=c2))

            i += 1






from PySide6.QtCore import Signal, Slot, Qt

from PySide6.QtWidgets import (
    QToolBox,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QVBoxLayout,
    QWidget,
    QMenu,
    QMessageBox,
    QPushButton, QInputDialog, QDialog
)
from .aircraft_info import aircraft_info
from .component_edit import component_edit
from .mission_profile_edit import mission_profile_edit
from .propulsion_input import propulsion_input
from .subcomponent_edit import subcomponent_edit
from .useful_load_edit import useful_load_edit
from .analysis_list import analysis_list

subcomponent_types = {'Nose Landing Gear': 'nose_landing_gear',
                      'Main Landing Gear': 'main_landing_gear',
                      'Air Conditioning': 'air_conditioning' ,
                      'Anti Icing': 'anti_ice',
                      "Auxiliary Power Unit (APU)": 'apu',
                      'Avionics': 'avionics',
                      'Electronics': 'electronics',
                      'Flight Controls': 'flight_controls',
                      'Furnishings': 'furnishings',
                      'Hydraulics': 'hydraulics',
                      'Instruments': 'instruments'}

useful_load_types = {'Passengers': 'passengers',
                      'Flight Attendants': 'flight_attendants',
                      'Pilots': 'pilots',
                      'Cargo': 'cargo',
                     'Fuel': 'fuel'}

mission_profile_types = {'Takeoff': 'takeoff',
                      'Climb': 'climb',
                      'Cruise': 'cruise',
                      'Descent': 'descent',
                        'Loiter': 'loiter',
                         'Weight Drop': "weight_drop",
                     'Landing': 'landing'}


class ToolBox(QToolBox):
    """
    Toolbox on the left side of the screen that controls all data input
    """

    component_changed = Signal(str)
    component_selected = Signal(str)
    component_removed = Signal(str)
    component_renamed = Signal(str, str)
    selected_component = ''

    # initiate all toolbox features
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.aircraft = parent.aircraft
        # Aircraft information
        label = QLabel()
        aircraft_input = aircraft_info(self)
        self.addItem(aircraft_input, "Aircraft Information")

        # Component list - double click to bring up popup to edit component
        self.component_list = QListWidget()

        self.populate_component_list()

        self.addItem(self.component_list, "Components")
        self.component_list.itemDoubleClicked.connect(self.handleComponentDoubleClicked)
        self.component_list.itemClicked.connect(self.handleComponentClicked)

        # Useful Load List
        useful_load_container = QWidget()
        layout = QVBoxLayout()
        useful_load_container.setLayout(layout)
        self.useful_load_list = QListWidget()
        for item in useful_load_types.keys():
            self.useful_load_list.addItem(item)
        self.useful_load_list.itemDoubleClicked.connect(self.handleUsefulLoadClicked)
        layout.addWidget(self.useful_load_list)

        self.useful_load_reset = QPushButton('Set Useful Load to Default Values')
        self.useful_load_reset.clicked.connect(self.reset_useful_load)
        layout.addWidget(self.useful_load_reset)

        self.addItem(useful_load_container, "Useful Load")

        self.setup_mission_profile_section()

        # Subcomponent list
        self.subcomponent_container = QWidget(self)
        layout = QVBoxLayout()
        self.subcomponent_container.setLayout(layout)
        self.subcomponent_list = QListWidget()
        for comp in subcomponent_types.keys():
            item = QListWidgetItem(comp)
            item.setFlags(item.flags() | Qt.ItemIsUserCheckable)
            item.setCheckState(Qt.Checked)
            self.subcomponent_list.addItem(item)
        self.subcomponent_list.itemChanged.connect(self.handleSubcomponentChecked)
        layout.addWidget(self.subcomponent_list)
        self.subsystem_reset = QPushButton('Set Subsystem Parameters to Default Values')
        self.subsystem_reset.clicked.connect(self.reset_subsystems)
        layout.addWidget(self.subsystem_reset)


        self.addItem(self.subcomponent_container, 'Subsystems')
        self.subcomponent_list.itemDoubleClicked.connect(self.handleSubcomponentSelected)


        self.propulsion = propulsion_input(self)
        self.propulsion.n_engines_changed.connect(self.handle_n_engines_changed)
        self.addItem(self.propulsion, 'Propulsion')

        self.analyses = analysis_list(self)
        self.addItem(self.analyses, "Analysis")

    # Tab 5: Mission Profile Section (new)
    def setup_mission_profile_section(self):
        mission_profile_container = QWidget()
        layout = QVBoxLayout()
        mission_profile_container.setLayout(layout)

        # List for mission profile
        self.mission_profile_list = QListWidget()
        for seg in self.aircraft.mission.mission_profile:
            self.mission_profile_list.addItem(seg.title)
        self.mission_profile_list.itemDoubleClicked.connect(self.handleMissionProfileClicked)
        layout.addWidget(self.mission_profile_list)

        # Reset button TODO: change this for mission profile stuff
        self.mission_profile_reset = QPushButton('Set Mission Profile to Default Values')
        self.mission_profile_reset.clicked.connect(self.reset_mission_profile)
        layout.addWidget(self.mission_profile_reset)

        self.addItem(mission_profile_container, "Mission Profile")

    # Creates a dialog box to input component characteristics
    def handleComponentDoubleClicked(self, item):

        pop = component_edit(item.text(), self)
        pop.component_changed.connect(self.component_changed)
        pop.title_changed.connect(self.handle_title_changed)
        pop.exec()

    def handleMissionProfileClicked(self, item):
        """Opens a pop-up dialog to edit the selected mission profile segment."""
        phase = item.text()
        popup = mission_profile_edit(self, phase)
        if popup.exec() == QDialog.Accepted:
            self.refresh_mission_profile_list()

    def refresh_mission_profile_list(self):
        self.mission_profile_list.clear()
        for seg in self.aircraft.mission.mission_profile:
            self.mission_profile_list.addItem(seg.title)

    def handleUsefulLoadClicked(self, item):
        popup = useful_load_edit(self, useful_load_types[item.text()])
        popup.exec()

    # Sends Signal to highlight selected Component
    @Slot(str)
    def handleComponentClicked(self, item):
        selected_component = item.text()
        # selected_component = self.component_list.selectedItems()[0].text()
        if selected_component == self.selected_component:
            self.component_list.clearSelection()
            selected_component = ''
        self.component_selected.emit(selected_component)
        self.selected_component = selected_component

    # Opens subcomponent dialogue
    def handleSubcomponentSelected(self, item):
        popup = subcomponent_edit(item.text(), self)
        popup.exec()

    # TODO Add functionality to this
    def handleSubcomponentChecked(self, item):
        if item.checkState() == Qt.Checked:
            title = subcomponent_types[item]


    # Sends Signal to change the title and update the component list
    @Slot(str, str)
    def handle_title_changed(self, old_title, new_title):
        if old_title:
            self.component_renamed.emit(old_title, new_title)
            # Find index in list to remove
            for index in range(self.component_list.count()):
                if self.component_list.item(index).text().lower() == old_title.lower():
                    self.component_list.item(index).setText(new_title)

    def contextMenuEvent(self, event):
        index = self.currentIndex()
        # Component List
        if index == 1:
            # Add component and Remove Component Options
            context_menu = QMenu(self)
            add_component = context_menu.addMenu("Add Component")
            add_component.addAction('Wing')
            add_component.addAction('Fuselage')
            add_component.addAction('Horizontal Stabilizer')
            add_component.addAction('Vertical Stabilizer')
            add_component.addAction('Nacelle')
            add_component.triggered.connect(self.handleAddComponent)

            remove_component = context_menu.addMenu("Remove Component")
            for comp in self.parent.aircraft.aero_components.keys():
                item = remove_component.addAction(comp)
                if comp.lower() == 'main wing' or comp.lower() == 'fuselage':
                    item.setEnabled(False)
            remove_component.triggered.connect(self.handleRemoveComponent)

            context_menu.exec(event.globalPos())

            # added
        elif index == 3:
            context_menu = QMenu(self)
            # add_segment = context_menu.addAction("Add Mission Profile Segment")
            # remove_segment = context_menu.addAction("Remove Selected Segment")
            #
            # add_segment.triggered.connect(self.handleAddMissionProfileSegment)
            # remove_segment.triggered.connect(self.handleRemoveMissionProfileSegment)
            #
            # context_menu.exec(event.globalPos())
            add_segment = context_menu.addMenu("Add Mission Profile Segment")
            for seg in mission_profile_types.keys():
                add_segment.addAction(seg)
            add_segment.triggered.connect(self.handleAddMissionProfileSegment)
            remove_segment = context_menu.addMenu("Remove Selected Segment")
            for seg in self.aircraft.mission.mission_profile:
                remove_segment.addAction(seg.title)
            remove_segment.triggered.connect(self.handleRemoveMissionProfileSegment)
            context_menu.exec(event.globalPos())

    def handleAddComponent(self, item):
        component = item.text().lower()
        if component == 'horizontal stabilizer':
            component = 'horizontal'
        if component == 'vertical stabilizer':
            component = 'vertical'
        pop = component_edit(component, self, new_component=True)
        pop.component_changed.connect(self.component_changed)
        pop.title_changed.connect(self.handle_title_changed)
        pop.exec()
        # Add the component to the component list
        self.populate_component_list()

    def populate_component_list(self):
        self.component_list.clear()
        for comp in self.aircraft.aero_components.values():
            self.component_list.addItem(comp.title)


    def handleRemoveComponent(self, item):
        component = item.text()

        def confirm_delete():
            msg = QMessageBox()
            msg.setWindowTitle('Confirm Delete')
            msg.setIcon(QMessageBox.Question)
            msg.setText(f'Are you sure you want to delete the component {component}')
            msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            msg.setDefaultButton(QMessageBox.No)
            result = msg.exec()
            return result == QMessageBox.Yes

        if confirm_delete():
            self.parent.aircraft.remove_component(component)
            self.component_changed.emit(component)

            # Find index in list to remove
            for index in range(self.component_list.count()):
                if self.component_list.item(index).text().lower() == component.lower():
                    self.component_list.takeItem(index)
                    return

    def handle_n_engines_changed(self, n_engines):
        pop = component_edit('Nacelle', self)
        pop.component_changed.connect(self.component_changed)
        pop.title_changed.connect(self.handle_title_changed)
        pop.exec()

    def handleAddMissionProfileSegment(self, item):
        segment_type = item.text().lower()

        popup = mission_profile_edit(self, segment_type.capitalize())
        if popup.exec() == QDialog.Accepted:
            self.refresh_mission_profile_list()


    def handleRemoveMissionProfileSegment(self):
        selected_item = self.mission_profile_list.currentItem()
        if selected_item:
            segment_name = selected_item.text()
            confirm = QMessageBox.question(
                self, "Confirm Delete",
                f"Are you sure you want to delete the mission profile segment '{segment_name}'?",
                QMessageBox.Yes | QMessageBox.No, QMessageBox.No
            )
            if confirm == QMessageBox.Yes:
                mission_profile_types.pop(segment_name, None)
                self.mission_profile_list.takeItem(self.mission_profile_list.row(selected_item))

    def reset_mission_profile(self):  # CHANGED
        confirm = QMessageBox()
        confirm.setWindowTitle('Confirm Mission Profile Reset')
        confirm.setIcon(QMessageBox.Question)
        confirm.setText('Are you sure you want to reset the mission profile to default values?')
        confirm.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        confirm.setDefaultButton(QMessageBox.No)
        confirmed = confirm.exec()

        if confirmed == QMessageBox.Yes:
            self.mission_profile_list.clear()
            for item in mission_profile_types.keys():
                self.mission_profile_list.addItem(item)

    def reset_subsystems(self):
        fuse = self.parent.aircraft.aero_components['Fuselage']

        confirm = QMessageBox()
        confirm.setWindowTitle('Confirm Subsystem Reset')
        confirm.setIcon(QMessageBox.Question)
        confirm.setText(f'Are you sure you want to reset the subsystems to default values')
        confirm.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        confirm.setDefaultButton(QMessageBox.No)
        confirmed = confirm.exec()

        if not confirmed:
            return

        for comp in self.parent.aircraft.subsystems.components.values():
            comp.cg = cg_default(comp.title, fuse)

        self.parent.aircraft.set_weight(self.parent.aircraft.weight_takeoff)

    def reset_useful_load(self):
        length = self.parent.aircraft.aero_components['Fuselage'].length

        confirm = QMessageBox()
        confirm.setWindowTitle('Confirm Subsystem Reset')
        confirm.setIcon(QMessageBox.Question)
        confirm.setText(f'Are you sure you want to reset the subsystems to default values')
        confirm.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        confirm.setDefaultButton(QMessageBox.No)
        confirmed = confirm.exec()

        if not confirmed:
            return

        useful_load = self.parent.aircraft.useful_load
        useful_load.cg_fuel = [65 * length / 123.3, 0 , 0]
        useful_load.cg_cargo = [54 * length / 123.3, 0, 0]
        useful_load.cg_passengers = [58 * length / 123.3, 0, 0]
        useful_load.cg_flight_attendants = [9 * length /123.3, 0, 0]
        useful_load.pilots = [9 * length /123.3, 0, 0]
        self.parent.aircraft.set_weight(self.parent.aircraft.weight_takeoff)

def cg_default(component, fuse):
    r = fuse.height / 2
    l = fuse.length
    component_cg = {'nose_landing_gear': [15, 0, -6],
                        'main_landing_gear': [69.3, 0, -6],
                        'air_conditioning': [62, 0, 0],
                        'anti_ice': [67, 0, 0],
                        'apu': [121, 0, 0],
                        'avionics': [9, 0, 0],
                        'electronics': [12, 0, 0],
                        'flight_controls': [0, 0, 0],
                        'furnishings': [56, 0, 0],
                        'hydraulics': [62, 0, 0],
                        'instruments': [9, 0, 0],
                      }
    cg = component_cg[component]
    cg[0] *= l / 123.31
    cg[1] *= r / 13.17
    cg[2] *= r / 13.17

    return cg


import logging

from .component import Component
from .aerobodies.wing import Wing
import numpy as np
logger = logging.getLogger(__name__)

# Contains component classes for all various subsystems
class Subsystems:


    def __init__(self, params):
        """
        Initializes all declared subsystems and sets parameters

        :param <dict> params: list of parameters to edit
        """
        """
            Contains weight properties for all supported subsystems
            """
        self.weight = 0
        self.cg = 0
        self.inertia = [0, 0, 0]

        self.parameters = {}
        self.components = {}
        self.parameters = params
        for name, cg in params['subsystems'].items():
            self.components[name] = Component({'title': name, 'cg': cg})

        # TODO take this out
        if not 'w_avionics' in self.parameters:
            self.parameters['w_avionics'] = 1200

    def set_subsystem_weights(self, aircraft, wdg):
        """
        Calculates weight, cg, and moments of inertia for all subsystems

        :param object aircraft: the aircraft which the weight, cg, and inertia calculations are done.

        :return weight of subsystems
        :rtype float
        """
        self.weight = 0
        self.inertia = [0, 0, 0]
        # Generate all components
        for name, comp in self.components.items():
            name = name.lower()

            if hasattr(self, name) and callable((getattr(self, name))):
                fun = getattr(self, name)
                self.weight += fun(aircraft, wdg, comp)
            else:
                logging.warning(f'Subsystem {name} does not exist')

            # Set inertia value, add to overall inertia
            comp.inertia = [comp.weight * x for x in comp.cg]
            self.inertia = [i1 + i2 for i1, i2 in zip(self.inertia, comp.inertia)]
        self.cg = [i / self.weight for i in self.inertia]
        return self.weight

    def nose_landing_gear(self, aircraft, wdg, comp):
        """
        Calculates weight of nose landing gear

        :param object aircraft: aircraft which the nose landing gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for nose landing gear.

        :return component weight
        :rtype float

        """
        params = self.parameters

        if 'w_nose_landing_gear' in params:
            comp.weight = params['w_nose_landing_gear']
        else:
            if aircraft.aircraft_type == 'transport':
                params = self.parameters
                # Raymer method (transport)
                if 'knp' in params:
                    knp = params['knp']
                else:
                    knp = 1
                w_raymer = .032 * knp * wdg ** .646 * aircraft.ultimate_load ** .2 * comp.cg[0] ** .5 * \
                           params['n_nose_wheels']

                # Torenbeek method (transport)
                if aircraft.aero_components['Main Wing'].zle > 0:
                    kgr = 1.08
                else:
                    kgr = 1.0
                ag = 25
                bg = 0
                cg = .0024
                dg = 0
                w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)

                comp.weight = .5 * (w_raymer + w_torenbeek)

            elif aircraft.aircraft_type == 'general_aviation':
                # Raymer
                w_raymer = .125 * (aircraft.ultimate_load * wdg) ** .566 * (comp.cg[0] / 12) ** .845
                # Torenbeek
                # if aircraft.aero_components['Main Wing'].zle > 0:
                #     kgr = 1.08
                # else:
                #     kgr = 1.0
                # ag = 25
                # bg = 0
                # cg = .0024
                # dg = 0
                # w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
                comp.weight = w_raymer  #* .2 + w_torenbeek * .8


        self.components[comp.title] = comp
        return comp.weight

    def main_landing_gear(self, aircraft, wdg, comp):
        """
        Sets main landing gear weight

        :param object aircraft: aircraft which the main landing gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for main landing gear.

        :return component weight
        :rtype float
        """
        params = self.parameters

        if 'w_main_landing_gear' in params:
            comp.weight = params['w_main_landing_gear']
        else:
            if aircraft.aircraft_type == 'transport':
                # Torenbeek method
                if aircraft.aero_components['Main Wing'].zle > 0:
                    kgr = 1.08
                else:
                    kgr = 1.0
                ag = 20
                bg = .1
                cg = .019
                dg = 0
                w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
                comp.weight = w_torenbeek

            elif aircraft.aircraft_type == 'general_aviation':
                ultimate_load_landing = aircraft.ultimate_load
                w_mlg = wdg  # landing weight
                length_mlg = comp.cg[0] # landing gear length (fix?)
                # raymer method
                w_raymer = .095 * (ultimate_load_landing * w_mlg) ** .768 * (length_mlg / 12) ** .409
                # NASA method
                w_nasa = .0117 * w_mlg ** .95 * length_mlg

               # torenbeek (need to make sure constants are correct)
                #if aircraft.aero_components['Main Wing'].zle > 0:
                #     kgr = 1.08
                # else:
                #     kgr = 1.0
                # ag = 20
                # bg = .1
                # cg = .019
                # dg = 0
                # w_torenbeek = kgr * (ag + bg * wdg ** .75 + cg * wdg + dg * wdg ** 1.5)
                # print(w_torenbeek)
                #comp.weight = (w_raymer* .2 + w_nasa * .2 + w_torenbeek * .6)
                #comp.weight =( w_raymer + w_nasa ) / 2
                comp.weight = w_raymer


        return comp.weight

    def air_conditioning(self, aircraft, wdg, comp):
        """
        Calculates weight of air conditioning
        """
        # Uses nasa flops method (same for general and transport)

        if 'w_air_conditioning' in self.parameters:
            comp.weight = self.parameters['w_air_conditioning']
        else:

            if 'Fuselage' in aircraft.aero_components.keys():
                l = aircraft.aero_components['Fuselage'].length
                w = aircraft.aero_components['Fuselage'].width
                h = aircraft.aero_components['Fuselage'].height

                w_nasa = 3.2 * (l * w * h) ** .6 + (9 * aircraft.mission.n_passengers ** .83) * aircraft.mission.max_mach + \
                         .075 * self.parameters['w_avionics']
                comp.weight = w_nasa

            if aircraft.aircraft_type == 'general_aviation':
                 #Raymer method (experimental)
                 w_raymer = .7 * (.265 * wdg ** .52 * aircraft.mission.n_passengers ** .68 * self.parameters['w_avionics'] ** .17 * aircraft.mission.max_mach ** .08)

                 comp.weight = .5 * (w_raymer + w_nasa)


        return comp.weight

    def anti_ice(self, aircraft, wdg, comp):
        """
        Calculate weight of anti ice components

        :param object aircraft: aircraft which the anti-ice gear belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for anit-ice gear.

        :return component weight
        :rtype float
        """
        if 'w_anti_ice' in self.parameters:
            comp.weight = self.parameters['w_anti_ice']
        else:
            # Raymer method
            w_raymer = .002 * wdg

            # find nacelle group
            for compon in aircraft.aero_components.values():
                if compon.component_type == 'Nacelle':
                    nacelle = compon

            try:
                d = nacelle.diameter * 3 / 5
            except:
                return 0

            # Nasa flops method
            b = aircraft.aero_components['Main Wing'].span
            sweep = aircraft.aero_components['Main Wing'].sweep
            l = aircraft.aero_components['Fuselage'].length
            w_nasa = b / np.cos(sweep) + (3.8 * d * aircraft.n_engines * 3) + 1.5 * l

            comp.weight = .5 * (w_nasa + w_raymer)
        return comp.weight

    def apu(self, aircraft, wdg, comp):
        """
        Calculate auxiliary power unit weight using torenbeek and nasa methods.

        :param object aircraft: aircraft  which the apu belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the apu.

        :return component weight
        :rtype float
        """
        if 'w_apu' in self.parameters:
            comp.weight = self.parameters['w_apu']
        else:

            if aircraft.aircraft_type == 'transport':
                # Torenbeek method
                w_torenbeek = .009 * wdg

                # nasa flops method
                d = aircraft.aero_components['Fuselage'].diameter
                l = aircraft.aero_components['Fuselage'].length
                w_nasa = 54 * (d * l) ** .3 + 5.4 * aircraft.mission.n_passengers ** .9

                comp.weight = .5 * (w_nasa + w_torenbeek)
            else:
                comp.weight = 0
                return 0

        return comp.weight


    def avionics(self, aircraft, wdg, comp):
        """
        Set avionics weight using raymer, torenbeek, and nasa methods

        :param object aircraft: aircraft which the avionics belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the avionics.

        :return component weight
        :rtype float
        """

        if 'w_avionics' in self.parameters:
            comp.weight = self.parameters['w_avionics']
        else:
            if aircraft.aircraft_type == 'transport':
                # Raymer method
                w_raymer = 1.73 * self.parameters['w_avionics'] ** .983

                # Torenbeek method
                w_torenbeek = .575 * (wdg - aircraft.mission.w_fuel) ** .556 * aircraft.mission.design_range ** .25

                # nasa method
                d = aircraft.aero_components['Fuselage'].diameter
                l = aircraft.aero_components['Fuselage'].length
                w_nasa = 15.8 * aircraft.mission.design_range ** .1 * aircraft.mission.n_pilots ** .7 * (l * d) ** .43

                comp.weight = (w_nasa + w_torenbeek + w_raymer) / 3


            elif aircraft.aircraft_type == 'general_aviation':
                # Raymer method
                w_raymer = 2.117 * self.parameters['w_avionics'] ** .933

                # # nasa method (same as transport)
                # d = aircraft.aero_components['Fuselage'].diameter
                # l = aircraft.aero_components['Fuselage'].length
                # w_nasa = 15.8 * aircraft.mission.design_range ** .1 * aircraft.mission.n_pilots ** .7 * (l * d) ** .43
                #comp.weight = (w_raymer + w_nasa) / 2
                comp.weight = w_raymer


        return comp.weight

    def electronics(self, aircraft, wdg, comp):
        """
        Calculates weight of all electronics on aircraft

        :param object aircraft: aircraft which the electronic belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the electronics.

        :return component weight
        :rtype float
        """

        if 'w_electronics' in self.parameters:
            comp.weight = self.parameters['w_electronics']
        else:
            # Only uses nasa method (same for both general aviation and transport aircrafts)
            if aircraft.aircraft_type == 'transport':
                d = aircraft.aero_components['Fuselage'].diameter
                l = aircraft.aero_components['Fuselage'].length
                nflcr = aircraft.mission.n_pilots + aircraft.mission.n_flight_attendants
                npass = aircraft.mission.n_passengers
                w_nasa = 92 * l ** .4 * d ** .14 * aircraft.n_engines ** .69 * (1 + .044 * nflcr + .0015 * npass)
                comp.weight = w_nasa

            elif aircraft.aircraft_type == 'general_aviation':
                # raymer
                w_avionics = self.parameters['w_avionics']
                w_fs = aircraft.aero_components['Nacelle'].weight_fuel_system
                comp.weight = 12.57 * (w_fs + w_avionics)**.51
        return comp.weight

    def flight_controls(self, aircraft, wdg, comp):
        """
        Calculates weight of flight controls.

        :param object aircraft: aircraft which the flight controls belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the flight controls.

        :return component weight
        :rtype float
        """
        if 'w_flight_controls' in self.parameters:
            comp.weight = self.parameters['w_flight_controls']
        else:
            scs = 0 # Control surface area
            scs_i = 0   # used to find center of gravity
            for compon in aircraft.aero_components.values():
                if hasattr(compon, 'area') and hasattr(compon, 'control_surface_ratio'):
                    scs += compon.area * compon.control_surface_ratio
                    scs_i += compon.area * compon.control_surface_ratio * \
                             (compon.xle + .25* compon.span * np.tan(compon.sweep))

            if scs > 0:
                comp.cg = [scs_i / scs, 0, 0]
            else:
                comp.cg = [0, 0, 0]

            # Raymer is innacurate
            # Torenbeek method (only factors in for transport aircrafts)
            w_torenbeek = .64 * wdg ** (2/3)

            # Nasa method

            if aircraft.aircraft_type == 'transport':
                w_nasa = 1.1 * aircraft.mach_cruise ** .52 * scs ** .6 * wdg ** .32
                comp.weight = .5 * (w_nasa + w_torenbeek)
            elif aircraft.aircraft_type == 'general_aviation':
                t0 = 519  # Rankine constant
                L = 0.003575  # temp lapse rate (R/ft)
                R = 1716.5  # Gas Constant
                t = t0 - L * aircraft.h_cruise
                theta = t / t0
                delta = theta ** (32.2 / (L * R))
                q_dive = 1481.35 * delta * aircraft.mission.max_mach ** 2
                w_nasa = .404 * aircraft.aero_components['Main Wing'].area ** .317 * (wdg / 1000) ** .602 * aircraft.ultimate_load ** .525 * q_dive ** .345
                comp.weight = w_nasa


        return comp.weight

    def furnishings(self, aircraft, wdg, comp):
        """
        Calculates weight of all miscellaneous furnishings.

        :param object aircraft: aircraft which the furnishings belongs to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the furnishings.

        :return component weight
        :rtype float
        """

        if 'w_furnishings' in self.parameters:
            comp.weight = self.parameters['w_furnishings']
        else:
            if aircraft.aircraft_type == 'transport':
                # Only using nasa, raymer and torenbeek are both complex and inacurate (nasa is same for transport and gen. av.)
                nflcr = aircraft.mission.n_pilots = aircraft.mission.n_flight_attendants
                npass = aircraft.mission.n_passengers
                w = aircraft.aero_components['Fuselage'].width
                d = aircraft.aero_components['Fuselage'].diameter
                xlp = aircraft.aero_components['Fuselage'].length * 80 / 125
                comp.weight = 127 * nflcr + 44 * npass + 2.6 * xlp * (w + d)
            elif aircraft.aircraft_type == 'general_aviation':
                comp.weight = .0582 * wdg - 65
        return comp.weight

    def hydraulics(self, aircraft, wdg, comp):
        """
        Calculates weight of hydraulics.

        :param object aircraft: aircraft which the hydraulics belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the hydraulics.

        :return component weight
        :rtype float
        """

        if 'w_hydraulics' in self.parameters:
            comp.weight = self.parameters['w_hydraulics']
        else:
            # Raymer is bad for this one

            # Torenbeek method
            w_torenbeek = .012 * wdg

            # nasa method
            d = aircraft.aero_components['Fuselage'].diameter
            l = aircraft.aero_components['Fuselage'].length

            b = aircraft.aero_components['Main Wing'].span
            s = aircraft.aero_components['Main Wing'].area

            mach = aircraft.cruise_conditions.mach

            n_fuse_engine = 0
            n_wing_engine = 0
            for compo in aircraft.aero_components.values():
                if compo.component_type == 'Nacelle':
                    if compo.attachment == 'Main Wing':
                        n_wing_engine += 2
                    elif compo.attachment == 'Fuselage':
                        n_fuse_engine += 2

            w_nasa = .57 * ((d * l) + (.27 * s)) * (1 + .03 * n_wing_engine + .05 * n_fuse_engine) * mach ** .33

            comp.weight = .5 * (w_nasa + w_torenbeek)

            # this is what was here before but im not sure where I got it from
            # elif aircraft.aircraft_type == 'general_aviation':
            #     s_w = aircraft.aero_components['Main Wing'].area
            #     s_vt = aircraft.aero_components['Vertical Stabilizer'].area
            #     s_ht = aircraft.aero_components['Horizontal Stabilizer'].area
            #     h_tc = aircraft.aero_components['Horizontal Stabilizer'].tc
            #
            #     w_nasa = .6053 * (s_w + 1.44 * (s_ht/(2 + .387 * h_tc)+ s_vt))
            #
            #     comp.weight = w_nasa

        return comp.weight

    def instruments(self, aircraft, wdg, comp):
        """
        Calculates instruments weights.

        :param object aircraft: aircraft which the instruments belong to.
        :param int wdg: takeoff gross weight (lbs)
        :param object comp: component object for the instruments.

        :return component weight
        :rtype float
        """
        if 'w_instruments' in self.parameters:
            comp.weight = self.parameters['w_instruments']
        else:
            # Raymer method
            kr = 1.0
            ktp = 1.0
            l = aircraft.aero_components['Fuselage'].length
            d = aircraft.aero_components['Fuselage'].diameter
            b = aircraft.aero_components['Main Wing'].span
            nflcr = aircraft.mission.n_pilots
            nen = aircraft.n_engines
            w_raymer = 4.509 * kr * ktp * nflcr ** .541 * nen * (l + b) ** .5

            # torenbeek method
            w_torenbeek = nflcr * (15 + .032 * (wdg / 1000)) + nen * (5 + .006 * (wdg / 1000))

            # nasa method
            mach = aircraft.cruise_conditions.mach
            w_nasa = .48 * (l * d) ** .57 * mach ** .5 * (10 + 2.5 * nflcr + nen)

            if aircraft.aircraft_type == 'transport':
                comp.weight = (w_nasa + w_torenbeek + w_raymer) / 3
            elif aircraft.aircraft_type == 'general_aviation':
                comp.weight = w_nasa

        return comp.weight


import os

from PySide6.QtWidgets import QListWidget, QFileDialog, QDialog, QVBoxLayout, QTextEdit, QMessageBox
from ...reports import weights_report, mission_profile_report

class analysis_list(QListWidget):

    def __init__(self, parent):
        self.parent = parent
        super().__init__(parent)
        self.addItems(['Weights Analysis', 'Mission Profile Analysis'])
        self.itemDoubleClicked.connect(self.handleItemSelected)

    def handleItemSelected(self, item):
        selection = item.text()
        if selection == 'Weights Analysis':
            self.generate_weights_report()
        if selection == 'Mission Profile Analysis':
            self.generate_mission_profile_analysis()

    def generate_weights_report(self):
        aircraft = self.parent.parent.aircraft

        file_dialog = QFileDialog(self)
        file_dialog.exec()
        filename = file_dialog.selectedFiles()[0]

        weights_report(aircraft, filename=filename)

        with open(filename, 'r') as file:
            report = file.read()

        report_popup = QDialog()
        report_popup.setWindowTitle('Weights Report')
        layout = QVBoxLayout(self)
        text_edit = QTextEdit(self)
        text_edit.setReadOnly(True)
        layout.addWidget(text_edit)
        report_popup.setLayout(layout)
        text_edit.setPlainText(report)
        report_popup.exec()

    def generate_mission_profile_analysis(self):
        aircraft = self.parent.parent.aircraft

        # Check if AVL is in the right folder
        wd = os.getcwd()
        if not os.path.exists(os.path.join(wd, 'avl.exe')):
            error = QMessageBox()
            error.setIcon(QMessageBox.Critical)
            error.setWindowTitle('Error')
            error.setText("Error: AVL file 'avl.exe' not found in current working directory")
            error.setInformativeText(f"Current working directory: {wd}")
            error.exec()
            return

        file_dialog = QFileDialog(self)
        file_dialog.exec()
        filename = file_dialog.selectedFiles()[0]
        aircraft.mission.run_case()
        mission_profile_report(aircraft, filename=filename)

        with open(filename, 'r') as file:
            report = file.read()

        report_popup = QDialog()
        report_popup.setWindowTitle('Weights Report')
        layout = QVBoxLayout(self)
        text_edit = QTextEdit(self)
        text_edit.setReadOnly(True)
        layout.addWidget(text_edit)
        report_popup.setLayout(layout)
        text_edit.setPlainText(report)
        report_popup.exec()
